import math
from typing import Union
from pathlib import Path
from planar_magnetics.geometry import Arc, Polygon, Point
from planar_magnetics.smoothing import smooth_polygon
from planar_magnetics.utils import dcr_of_annulus
from planar_magnetics.materials import Conductor, COPPER
from planar_magnetics.windings.windings import Winding


class Spiral(Winding):
    """An optimized spiral multi-turn winding on a single layer

    Args:
        inner_radius: The inner radius of the spiral specified in mm
        outer_radius: The outer radius of the spiral specified in mm
        num_turns: The number of turns in the spiral
        spacing: The inter-turn spacing specified in mm
        layer: A string representing what layer the spiral is on
        radius: The radius used to smooth out the corners of the shape
        at: The center of the spiral

    Attributes:
        polygon: The final shape of the spiral
    """

    def __init__(
        self,
        inner_radius: float,
        outer_radius: float,
        num_turns: int,
        spacing: float,
        layer: str = "F.Cu",
        radius: float = 0,
        at: Point = Point(0, 0),
    ):
        # calculate optimal turn radii using equation 10 from Conceptualization and Analysis of a
        # Next-Generation Ultra-Compact 1.5-kW PCB-Integrated Wide-Input-Voltage-Range 12V-Output
        # Industrial DC/DC Converter Module
        # https://www.pes-publications.ee.ethz.ch/uploads/tx_ethpublications/7_electronics-10-02158_FINAL_Knabben.pdf
        inverse_num_turns = 1 / num_turns
        radii = [
            (inner_radius ** (num_turns - i) * outer_radius ** i) ** inverse_num_turns
            for i in range(num_turns)
        ]

        # verify that the minimum trace width is greater than 2 x min radius
        min_trace_width = (
            radii[1] - radii[0] - spacing
            if num_turns > 1
            else outer_radius - inner_radius
        )
        assert (
            min_trace_width > 2 * radius
        ), f"This spiral requires a min trace width of {1e3*min_trace_width}mm, which is less than 2 x radius"

        # create the arcs for the inner turns
        angle = math.acos(1 - spacing / radii[0])
        arcs = [Arc(at, radii[0], -math.pi + angle, math.pi)]
        for r0, r1 in zip(radii[0:-1], radii[1:]):
            angle = math.acos(spacing / r1 + r0 * (1 - spacing / r0) / r1)
            arc = Arc(at, r1, -math.pi + angle, math.pi)
            arcs.append(arc)

        # create the outermost arc
        a0 = math.acos(
            spacing / outer_radius
            + radii[-1] * (1 - spacing / radii[-1]) / outer_radius
        )
        a1 = math.acos(radii[-1] * (1 - spacing / radii[-1]) / outer_radius)
        arc = Arc(at, outer_radius, math.pi + a0, -math.pi + a1)
        arcs.append(arc)

        # create the outer arcs of the other turns
        for r0, r1 in zip(radii[-2::-1], radii[-1:0:-1]):
            angle = math.acos((r0 - spacing) / (r1 - spacing))
            arc = Arc(at, r1 - spacing, math.pi, -math.pi + angle)
            arcs.append(arc)

        polygon = Polygon(arcs, layer)

        # Add smoothing if a positive radius is provided
        if radius > 0:
            polygon = smooth_polygon(polygon, radius)

        self.polygon = polygon

        self.inner_radii = radii
        self.outer_radii = [r - spacing for r in radii[1:]] + [outer_radius]

    def estimate_dcr(
        self, thickness: float, termperature: float = 25, material: Conductor = COPPER
    ):
        """Estimate the DC resistance of the winding

        This function will estimate the DC resistance of the winding by calculating the estimated
        dc resistance of each turn and adding the estimated inter-turn via resistance 
        
        Args:
            thickness: The copper thickness of the layer
            temperature: The temperature of the winding in degrees Celcius
            material: The material used for the winding

        Returns:
            float: An estimation of the DC resistance in ohms
        """
        # calculate the material resistivity
        rho = material.get_resistivity(termperature)

        # sum the resistance of each turn
        resistance = 0
        for r0, r1 in zip(self.inner_radii, self.outer_radii):
            resistance += dcr_of_annulus(thickness, r0, r1, rho)

        return resistance


if __name__ == "__main__":

    from planar_magnetics.creepage import calculate_creepage
    from planar_magnetics.utils import weight_to_thickness

    # create a spiral inductor
    spiral = Spiral(
        at=Point(110e-3, 110e-3),
        inner_radius=6e-3,
        outer_radius=12e-3,
        num_turns=1,
        spacing=calculate_creepage(500, 1),
        layer="F.Cu",
        radius=0.3e-3,
    )

    # estimate the dc resistance of this spiral assuming 2 oz copper
    dcr = spiral.estimate_dcr(thickness=weight_to_thickness(2))
    print(f"Estimated DCR of this spiral is {dcr*1e3} mOhms")

    # dispay a preview of the spiral from Python using matplotlib
    spiral.plot()

    # # export this to a DXF file
    # spiral.export_to_dxf("spiral.dxf")

    # get the KiCad S expression, which can be then be copy-pasted into a KiCAD footprint file and edited from the footprint editer
    print(spiral)



import math
import uuid
from planar_magnetics.materials import Ferrite, N96
from planar_magnetics.geometry import (
    Arc,
    Point,
    Polygon,
    TWO_PI,
    PI_OVER_TWO,
)


def calculate_core_extension(area: float, radius: float, opening_width: float) -> float:
    """Calculate the required core extension to have the same area as the center-post

    Args:
        area (float): The total required area of the outer legs
        radius (float): The radius of the edge of the outer legs

    Returns:
        float: The required extension
    """

    # first, calculate the area of a single leg with no extension
    start_angle = math.asin((opening_width / 2) / radius)
    end_angle = PI_OVER_TWO - start_angle

    x = math.sqrt(radius ** 2 - (opening_width / 2) ** 2)

    # start with the area of the square from the center to the corner
    min_leg_area = x ** 2

    # subtract the opening triangles
    min_leg_area -= x * opening_width / 2

    # subtract the arc
    min_leg_area -= ((end_angle - start_angle) / TWO_PI) * (math.pi * radius ** 2)

    # calculated the required extension area
    extension_area = area / 4 - min_leg_area

    # if the minimum leg area is already sufficient, then just return 0
    if extension_area <= 0:
        return 0

    # calculate the minimum width of the leg
    x = radius * (math.cos(start_angle) - math.cos(end_angle))

    # solve for the extension using the quadratic equasion
    extension = (-2 * x + math.sqrt(4 * x ** 2 + 4 * extension_area)) / 2

    return extension


class Core:
    """A magnetic core object

    Args:
        centerpost_radius: The centerpost radius in mm
        window_width: The winding window width in mm
        window_height: The winding window height in mm
        opening_width: The core termination opening widths in mm
        gap: The core cap in mm

    Attributes:
        centerpost_area: The area of the centerpost in square mm
        volume: The core volume in cubic mm
    """

    def __init__(
        self,
        centerpost_radius: float,
        window_width: float,
        window_height: float,
        opening_width: float = None,
        gap: float = 0,
    ):

        self.centerpost_radius = centerpost_radius
        self.window_width = window_width
        self.window_height = window_height

        # if the user does not specify a opening width for terminations, just use the window width
        if opening_width is None:
            self.opening_width = window_width
        else:
            self.opening_width = opening_width

        self.gap = gap

        # calculate core dimensions
        self.outerpost_radius = centerpost_radius + window_width

        # calculate the centerpost area
        self.centerpost_area = math.pi * self.centerpost_radius ** 2

        # calculate the start angle first outer post cutout leg
        start_angle = math.asin((self.opening_width / 2) / self.outerpost_radius)

        # calculate the end angle of the first outer post cutout leg
        end_angle = math.pi / 2 - start_angle

        # create polygons for the outer post cutouts
        extension = calculate_core_extension(
            area=self.centerpost_area,
            radius=self.outerpost_radius,
            opening_width=self.opening_width,
        )

        # Calculate the plate thickness so that the plate mating surface has the same area as the
        # centerpost area
        self.plate_thickness = self.centerpost_area / (
            2 * self.centerpost_radius * math.pi
        )

        self.width = 2 * (self.outerpost_radius * math.cos(start_angle) + extension)
        self.height = self.window_height + 2 * self.plate_thickness

    def get_coreloss(
        self, B: float, f: float, ferrite: Ferrite = N96, temperature: float = 25
    ):
        """Calculate the coreloss using steinmetz parameters
        """

        # This is not tested, so raise error
        raise NotImplementedError

        k, alpha, beta = ferrite.get_steinmetz_parameters(temperature)

        Pv = k * f ** alpha * B ** beta

        return Pv * self.volume

    def to_parts(
        self,
        freecad_path: str = "C:/Program Files/FreeCAD 0.19/bin",
        tol: float = 0.1,
        spacer_thickness: float = 0.0,
    ):

        # try and import the FreeCAD python extension
        try:
            import sys

            sys.path.append(freecad_path)
            import FreeCAD as cad
        except Exception:
            raise ImportError("You must have FeeCAD installed")
        import Part

        # create the center piece
        circle = Part.makeCircle(self.centerpost_radius)
        wire = Part.Wire(circle)
        disk = Part.Face(wire)
        centerpost = disk.extrude(cad.Vector(0, 0, self.window_height / 2))
        to_gap = self.centerpost_radius + self.window_width / 2 - self.gap / 2
        circle = Part.makeCircle(to_gap, cad.Vector(0, 0, self.window_height / 2))
        wire = Part.Wire(circle)
        disk = Part.Face(wire)
        topplate = disk.extrude(cad.Vector(0, 0, self.plate_thickness))
        centerpiece = centerpost.fuse(topplate)

        # create the top plate
        square = Part.makePlane(
            self.width,
            self.width,
            cad.Vector(-self.width / 2, -self.width / 2, self.window_height / 2,),
        )
        circle = Part.makeCircle(
            to_gap + self.gap, cad.Vector(0, 0, self.window_height / 2)
        )
        wire = Part.Wire(circle)
        disk = Part.Face(wire)
        topplate_face = square.cut(disk)
        topplate = topplate_face.extrude(cad.Vector(0, 0, self.plate_thickness))

        # create the legs
        opening_left_to_right = Part.makePlane(
            self.width,
            self.opening_width,
            cad.Vector(
                -self.width / 2, -self.opening_width / 2, self.window_height / 2,
            ),
        )
        opening_front_to_back = Part.makePlane(
            self.opening_width,
            self.width,
            cad.Vector(
                -self.opening_width / 2, -self.width / 2, self.window_height / 2,
            ),
        )
        circle = Part.makeCircle(
            self.outerpost_radius, cad.Vector(0, 0, self.window_height / 2),
        )
        wire = Part.Wire(circle)
        disk = Part.Face(wire)
        leg_face = square.cut(disk)
        legs_face = leg_face.cut(opening_left_to_right)
        legs_face = legs_face.cut(opening_front_to_back)
        legs = legs_face.extrude(cad.Vector(0, 0, -self.window_height / 2))

        # fuse the legs to the top plate
        topplate = topplate.fuse(legs)

        # create the spacer
        circle = Part.makeCircle(
            self.outerpost_radius - tol, cad.Vector(0, 0, self.window_height / 2),
        )
        wire = Part.Wire(circle)
        disk = Part.Face(wire)
        circle = Part.makeCircle(
            self.centerpost_radius + tol, cad.Vector(0, 0, self.window_height / 2),
        )
        wire = Part.Wire(circle)
        cutout = Part.Face(wire)
        washer = disk.cut(cutout)
        spacer = washer.extrude(cad.Vector(0, 0, -spacer_thickness))
        circle = Part.makeCircle(
            to_gap + self.gap - tol, cad.Vector(0, 0, self.window_height / 2),
        )
        wire = Part.Wire(circle)
        disk = Part.Face(wire)
        circle = Part.makeCircle(to_gap + tol, cad.Vector(0, 0, self.window_height / 2))
        wire = Part.Wire(circle)
        cutout = Part.Face(wire)
        washer = disk.cut(cutout)
        gap_spacer = washer.extrude(cad.Vector(0, 0, self.plate_thickness))
        spacer = spacer.fuse(gap_spacer)

        if not self.gap:
            top_half = centerpiece.fuse(topplate)
            top_half = top_half.removeSplitter()

            return {"core": top_half, "spacer": spacer}
        else:
            topplate = topplate.removeSplitter()
            centerpiece = centerpiece.removeSplitter()

            return {
                "core center": centerpiece,
                "core outer": topplate,
                "spacer": spacer,
            }

        # # bottom_half = top_half.mirror(cad.Vector(0, 0, 0), cad.Vector(0, 0, -1))

        # core = Part.makeCompound([top_half, bottom_half])

        # # scale part
        # core = core.scale(scale)

        # return core

    def to_step(
        self, filename: str, freecad_path: str = "C:/Program Files/FreeCAD 0.19/bin",
    ):
        """Export the core geometry to a step file
        """

        # try and import the FreeCAD python extension
        try:
            import sys

            sys.path.append(freecad_path)
            import FreeCAD as cad
        except Exception:
            raise ImportError("You must have FeeCAD installed")
        import Part

        parts = [part for part in self.to_parts().values() if part is not None]
        part = Part.makeCompound(parts)

        # top_half = parts[0]
        # for part in parts[1:]:
        #     top_half = top_half.fuse(part)

        # # bottom_half = top_half.mirror(cad.Vector(0, 0, 0), cad.Vector(0, 0, -1))
        part.exportStep(filename)

    def create_pcb_cutouts(self, center: Point = Point(0, 0), clearance: float = 0.5):
        """Generate cutout polygons"""

        # calculate the radius of the outer post cutouts
        outer_cutout_radius = self.outerpost_radius - clearance

        # calculate the start angle first outer post cutout leg
        start_angle = math.asin(
            (self.opening_width / 2 - clearance) / outer_cutout_radius
        )

        # calculate the end angle of the first outer post cutout leg
        end_angle = math.pi / 2 - start_angle

        # TODO: This could be a circle if it was supported
        centerpost = Polygon(
            [Arc(center, self.centerpost_radius + clearance, -math.pi, math.pi)],
            "Edge.Cuts",
            0.1,
            "none",
        )

        cutout_extension = (
            self.width / 2 + clearance - outer_cutout_radius * math.cos(start_angle)
        )

        # calculate the arcs
        arc = Arc(center, outer_cutout_radius, start_angle, end_angle)
        corner1 = arc.end + Point(0, cutout_extension)
        corner3 = arc.start + Point(cutout_extension, 0)
        corner2 = Point(corner3.x, corner1.y)
        leg1 = Polygon([arc, corner1, corner2, corner3], "Edge.Cuts", 0.1, "none")

        start_angle += math.pi / 2
        end_angle += math.pi / 2
        arc = Arc(center, outer_cutout_radius, start_angle, end_angle)
        corner1 = arc.end + Point(-cutout_extension, 0)
        corner3 = arc.start + Point(0, cutout_extension)
        corner2 = Point(corner1.x, corner3.y)
        leg2 = Polygon([arc, corner1, corner2, corner3], "Edge.Cuts", 0.1, "none")

        start_angle += math.pi / 2
        end_angle += math.pi / 2
        arc = Arc(center, outer_cutout_radius, start_angle, end_angle)
        corner1 = arc.end + Point(0, -cutout_extension)
        corner3 = arc.start + Point(-cutout_extension, 0)
        corner2 = Point(corner3.x, corner1.y)
        leg3 = Polygon([arc, corner1, corner2, corner3], "Edge.Cuts", 0.1, "none")

        start_angle += math.pi / 2
        end_angle += math.pi / 2
        arc = Arc(center, outer_cutout_radius, start_angle, end_angle)
        corner1 = arc.end + Point(cutout_extension, 0)
        corner3 = arc.start + Point(0, -cutout_extension)
        corner2 = Point(corner1.x, corner2.y)
        leg4 = Polygon([arc, corner1, corner2, corner3], "Edge.Cuts", 0.1, "none")

        cutouts = [centerpost, leg1, leg2, leg3, leg4]

        return cutouts


if __name__ == "__main__":

    core = Core(8.6, 6, 6, 3, 0.5)
    core.to_step("core.step")
